#include"dumberdowner.h"

#include<stdlib.h>
#include<assert.h>

// This is the dumbing down pass.
//
// Complex expressions are to be "broken down" into simpler ones until the AST
// can be trivially translated to the target architecture.
//
// This file along with CG is strictly for IA-32 and will fail for other
// architecture.

#include"x86.h"

/* Split away complex expression into a new local variable */
static AST *varify(AST *tlc, AST *chunk, AST *stmtPrev, AST *stmt, AST *e) {
	VarTableEntry *vte = malloc(sizeof(*vte));
	vte->kind = VARTABLEENTRY_VAR;
	vte->type = e->expression.type;
	vte->data.var.color = 0;
	vte->data.var.degree = 0;
	vte->data.var.priority = 0;
	vte->data.var.reachingDefs = NULL;
	
	// Add to var array
	tlc->chunk.vars = realloc(tlc->chunk.vars, sizeof(*tlc->chunk.vars) * (++tlc->chunk.varCount));
	tlc->chunk.vars[tlc->chunk.varCount - 1] = vte;
	
	// Alter AST
	
	ASTExprVar *ev[2];
	for(int i = 0; i < 2; i++) {
		ev[i] = malloc(sizeof(ASTExprVar));
		ev[i]->nodeKind = AST_EXPR_VAR;
		ev[i]->type = e->expression.type;
		ev[i]->thing = vte;
	}
	
	ASTStmtAssign *assign = malloc(sizeof(*assign));
	assign->nodeKind = AST_STMT_ASSIGN;
	assign->what = (AST*) ev[0];
	assign->to = e;
	
	if(stmtPrev) {
		stmtPrev->statement.next = (AST*) assign;
	} else {
		chunk->chunk.statementFirst = (AST*) assign;
	}
	assign->next = stmt;
	
	// Reset ud-chain
	
	vte->data.var.usedefFirst = NULL;
	vte->data.var.usedefLast = NULL;
	
	return (AST*) ev[1];
}

static AST *xopify(AST *tlc, AST *chunk, AST *stmtPrev, AST *stmt, AST *e) {
	return varify(tlc, chunk, stmtPrev, stmt, e);
}

static int dumben_chunk(AST *tlc, AST *chu) {
	AST *sPrev = NULL;
	AST *s = chu->chunk.statementFirst;
	
	int effective = 0;
	
	while(s) {
	
		if(s->nodeKind == AST_STMT_IF) {
			
			AST *e = s->stmtIf.expression;
			
			if(e->nodeKind == AST_EXPR_BINARY_OP && binop_is_comparison(e->exprBinOp.operator)) {
			
				if(is_xop(e->exprBinOp.operands[0]) == XOP_NOT_XOP) {
					e->exprBinOp.operands[0] = xopify(tlc, chu, sPrev, s, e->exprBinOp.operands[0]);
					effective = 1;
				}
				
				if(is_xop(e->exprBinOp.operands[1]) == XOP_NOT_XOP) {
					e->exprBinOp.operands[1] = xopify(tlc, chu, sPrev, s, e->exprBinOp.operands[1]);
					effective = 1;
				}
				
				if(is_xop(e->exprBinOp.operands[0]) == XOP_MEM && is_xop(e->exprBinOp.operands[1]) == XOP_MEM) {
					// Can't have two mems; put one in var
					
					e->exprBinOp.operands[1] = varify(tlc, chu, sPrev, s, e->exprBinOp.operands[1]);
					effective = 1;
				}
				
			} else {
				
				s->stmtIf.expression = varify(tlc, chu, sPrev, s, e);
				effective = 1;
				
			}
			
			effective |= dumben_chunk(tlc, s->stmtIf.then);
			
		} else if(s->nodeKind == AST_STMT_LOOP) {
			
			effective |= dumben_chunk(tlc, s->stmtLoop.body);
			
		} else if(s->nodeKind == AST_STMT_ASSIGN) {
			
			if(s->stmtAssign.what->nodeKind == AST_EXPR_UNARY_OP && s->stmtAssign.what->exprUnOp.operator == UNOP_DEREF
				&& s->stmtAssign.to->nodeKind == AST_EXPR_UNARY_OP && s->stmtAssign.to->exprUnOp.operator == UNOP_DEREF) {
				
				s->stmtAssign.to = varify(tlc, chu, sPrev, s, s->stmtAssign.to);
				effective = 1;
			}
			
			if(s->stmtAssign.to && s->stmtAssign.to->nodeKind == AST_EXPR_UNARY_OP && s->stmtAssign.to->exprUnOp.operator == UNOP_NEGATE && !ast_expression_equal(s->stmtAssign.what, s->stmtAssign.to->exprUnOp.operand)) {
				// Turn this:
				//  a = -b
				// into
				//  a = b
				//  a = -a
				
				AST *ev[2] = {
					ast_deep_copy(s->stmtAssign.what),
					ast_deep_copy(s->stmtAssign.what)
				};
				
				AST *negation = s->stmtAssign.to;
				
				s->stmtAssign.to = negation->exprUnOp.operand;
				negation->exprUnOp.operand = ev[0];
				
				AST *assign2 = malloc(sizeof(ASTStmtAssign));
				assign2->nodeKind = AST_STMT_ASSIGN;
				assign2->stmtAssign.what = ev[1];
				assign2->stmtAssign.to = negation;
				
				assign2->statement.next = s->statement.next;
				s->statement.next = assign2;
				
				effective = 1;
			}
			
			if(is_xop(s->stmtAssign.to) == XOP_NOT_XOP) {
				if(s->stmtAssign.to->nodeKind == AST_EXPR_UNARY_OP && s->stmtAssign.to->exprUnOp.operator == UNOP_DEREF) {
					s->stmtAssign.to->exprUnOp.operand = varify(tlc, chu, sPrev, s, s->stmtAssign.to->exprUnOp.operand);
					
					effective = 1;
				} else if(s->stmtAssign.to->nodeKind == AST_EXPR_BINARY_OP && s->stmtAssign.to->exprBinOp.operator < BINOP_SIMPLES && !ast_expression_equal(s->stmtAssign.what, s->stmtAssign.to->exprBinOp.operands[0])) {
					// Turn this:
					//  a = b op c
					// into
					//  a = b
					//  a = a op c
					
					AST *assign2 = malloc(sizeof(ASTStmtAssign));
					assign2->nodeKind = AST_STMT_ASSIGN;
					assign2->stmtAssign.what = ast_deep_copy(s->stmtAssign.what);
					assign2->stmtAssign.to = s->stmtAssign.to->exprBinOp.operands[0];
					
					sPrev->statement.next = assign2;
					assign2->statement.next = s;
					
					s->stmtAssign.to->exprBinOp.operands[0] = ast_deep_copy(s->stmtAssign.what);
					
					effective = 1;
				} else if(s->stmtAssign.to->nodeKind == AST_EXPR_BINARY_OP && !is_xop(s->stmtAssign.to->exprBinOp.operands[0])) {
					s->stmtAssign.to->exprBinOp.operands[0] = xopify(tlc, chu, sPrev, s, s->stmtAssign.to->exprBinOp.operands[0]);
				} else if(s->stmtAssign.to->nodeKind == AST_EXPR_BINARY_OP && !is_xop(s->stmtAssign.to->exprBinOp.operands[1])) {
					s->stmtAssign.to->exprBinOp.operands[1] = xopify(tlc, chu, sPrev, s, s->stmtAssign.to->exprBinOp.operands[1]);
				}
			}
			
		} else if(s->nodeKind == AST_STMT_EXPR && s->stmtExpr.expr->nodeKind == AST_EXPR_CALL) {
			
			int argCount = s->stmtExpr.expr->exprCall.what->expression.type->function.argCount;
			
			for(int i = 0; i < argCount; i++) {
				if(is_xop(s->stmtExpr.expr->exprCall.args[i]) == XOP_NOT_XOP) {
					s->stmtExpr.expr->exprCall.args[i] = xopify(tlc, chu, sPrev, s, s->stmtExpr.expr->exprCall.args[i]);
					effective = 1;
				}
			}
			
		}
		
		sPrev = s;
		s = s->statement.next;
	}
	
	return effective;
}

void dumben_go(AST* tlc) {
	while(dumben_chunk(tlc, tlc));
}
